from django.core.management.base import BaseCommand
import logging


class Command(BaseCommand):
    help = 'Update dynamic spider content e.g. permissions, content'

    def handle(self, *args, **options):
        from spkcspider.apps.spider.signals import update_dynamic
        self.log = logging.getLogger(__name__)
        for handler in self.log.handlers:
            self.log.removeHandler(handler)
        self.log.addHandler(logging.StreamHandler(self.stdout))
        results = update_dynamic.send_robust(self)
        for (receiver, result) in results:
            if isinstance(result, Exception):
                self.log.error(
                    "%s failed", receiver, exc_info=result
                )



# Generated by Django 2.1.7 on 2019-03-02 17:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0004_auto_20190225_2153'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dataverificationtag',
            name='verification_state',
            field=models.CharField(choices=[('pending', 'pending'), ('verified', 'verified'), ('invalid', 'invalid'), ('rejected', 'rejected')], default='pending', max_length=10),
        ),
    ]



__all__ = ("Command",)

import json

from django.core.management.base import BaseCommand
from django.apps import apps


class Command(BaseCommand):
    help = (
        "fix json fields in string format"
    )

    def fix_instance(self, instance, field):
        if isinstance(getattr(instance, field), str):
            try:
                setattr(
                    instance,
                    field,
                    json.loads(
                        getattr(instance, field).replace("'", "\"")
                        .replace("False", "false")
                        .replace("True", "true")
                    )
                )
                self.stdout.write("Fixed instance: {}".format(instance))
            except Exception as exc:
                self.stdout.write(str(exc))
                self.stdout.write("\n")

    def handle(self, **options):
        from jsonfield import JSONField as step
        JSONField = type(step())

        for model in apps.get_models():
            requires_hack = []
            for field in model._meta.get_fields():
                if isinstance(field, JSONField):
                    requires_hack.append(field.name)
            if requires_hack:
                for i in model.objects.all():
                    for field in requires_hack:
                        self.fix_instance(i, field)
                    try:
                        i.clean()
                        i.save()
                    except Exception as exc:
                        self.stdout.write(str(exc))
                        self.stdout.write("\n")



# Generated by Django 2.2 on 2019-04-04 21:16

import django.core.validators
from django.db import migrations, models

import spkcspider.apps.spider.validators


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20190331_1222'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='assignedcontent',
            name='fake_id',
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='description',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='name',
            field=models.CharField(blank=True, default='', max_length=255, validators=[
                spkcspider.apps.spider.validators.content_name_validator
            ]),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.CharField(db_index=True, help_text='Name of the component.<br/>Note: there are special named components with different protection types and scopes.<br/>Most prominent: "index" for authentication', max_length=255, validators=[django.core.validators.RegexValidator('^(\\w[\\w ]*\\w|\\w?)$')]),
        ),
    ]



__all__ = ["SpiderFiletsConfig"]

from django.apps import AppConfig

from django.db.models.signals import post_delete


def DeleteFilesCb(sender, instance, **kwargs):
    if instance.file:
        instance.file.delete(False)


class SpiderFiletsConfig(AppConfig):
    name = 'spkcspider.apps.spider_filets'
    label = 'spider_filets'
    verbose_name = 'spkcspider File and Text content'

    def ready(self):
        from .models import FileFilet
        post_delete.connect(
            DeleteFilesCb, sender=FileFilet,
            dispatch_uid="delete_files_filet"
        )



__all__ = ("Command",)

import datetime
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = (
        "Delete old/all auth tokens"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            '--days', action='store', dest='days', default=None,
            help='Delete tokens older than days',
        )
        parser.add_argument(
            '--oldest', action='store', dest='oldest', default=None,
            help='Delete oldest x tokens',
        )
        parser.add_argument(
            '--referrer', action='store', dest='referrer', default=None,
            help='Delete tokens of referrer',
        )
        parser.add_argument(
            '--anchor', action='store', dest='anchor', default=None,
            help=(
                'Delete tokens of anchor id:\n'
                '0/"component" for component,\n'
                '"all":  all tokens,\n'
                '"persist": persistent tokens'
            ),
        )

    def handle(self, oldest=None, days=None, anchor=None, **options):
        from spkcspider.apps.spider.models import AuthToken
        q = AuthToken.objects.all()

        if options['referrer']:
            q = q.filter(referrer__url=options["referrer"])
            if anchor == "all" or not anchor:
                pass
            elif anchor == "component":
                q = q.filter(persist=0)
            elif anchor == "persist":
                q = q.filter(persist__gte=0)
            else:
                q = q.filter(persist=int(anchor))
        elif anchor == "component":
            q = q.filter(persist=0)
        elif anchor == "persist":
            q = q.filter(persist__gte=0)
        elif anchor == "all":
            pass
        elif anchor:
            q = q.filter(persist=int(anchor))
        else:
            q = q.filter(persist=-1)
        if days:
            dt = datetime.datetime.now() - \
                datetime.timedelta(days=int(days))
            q.filter(created__lt=dt)
        if oldest:
            # create list with ids (liat for mysql required)
            ids = list(q.order_by("created").values_list(
                'id', flat=True
            )[:int(oldest)])
            # recreate Query
            q = AuthToken.objects.filter(id__in=ids)
        self.stdout.write("count: %s\n" % q.delete()[0])




import logging
from urllib.parse import urljoin

from django.db import models
from django.conf import settings
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string
from django.utils.translation import pgettext
from django.http import HttpResponsePermanentRedirect, JsonResponse

from jsonfield import JSONField

from spkcspider.apps.spider.helpers import get_hashob
from spkcspider.apps.spider.contents import BaseContent, add_content
from spkcspider.apps.spider.constants import VariantType
from spkcspider.apps.spider.conf import get_anchor_domain

logger = logging.getLogger(__name__)


# Create your models here.

_htest = get_hashob()
_htest.update(b"test")

_help_text_sig = _("""Signature of Identifier (hexadecimal-encoded)""")

_help_text_key = _(""""Public Key"-Content for signing identifier. It is recommended to use different keys for signing and encryption.""")  # noqa

ID_VERIFIERS = {

}


def valid_pkey_properties(key):
    _ = gettext
    if "PRIVAT" in key.upper():
        raise ValidationError(_('Private Key'))
    if key.strip() != key:
        raise ValidationError(_('Not trimmed'))
    if len(key) < 100:
        raise ValidationError(_('Not a key'))


@add_content
class PublicKey(BaseContent):
    expose_name = False
    expose_description = True
    appearances = [
        {"name": "PublicKey", "ctype": VariantType.unique.value}
    ]

    key = models.TextField(
        editable=True, validators=[valid_pkey_properties],
        help_text=_(
            "It is recommended to use different keys"
            "for signing and encryption"
        )
    )

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Public Key")

    def get_size(self):
        s = super().get_size()
        s += len(self.key)
        return s

    def get_info(self):
        ret = super().get_info()
        key = self.get_key_name()[0]
        h = get_hashob()
        h.update(key.encode("ascii", "ignore"))
        return "%shash=%s=%s\x1e" % (
            ret, settings.SPIDER_HASH_ALGORITHM.name, h.finalize().hex()
        )

    def get_key_name(self):
        # PEM key
        split = self.key.split("\n")
        if len(split) > 1:
            h = get_hashob()
            h.update(self.key.encode("ascii", "ignore"))
            return (h.finalize().hex(), None)

        # ssh key
        split = self.key.rsplit(" ", 1)
        if len(split) == 2 and "@" in split[1]:
            return split
        # other key
        return (self.key, None)

    def get_content_name(self):
        st = self.get_key_name()
        if st[1]:
            st = st[1]
        else:
            st = "{}...".format(st[0][:10])
        if len(self.associated.description) > 0:
            st = "{}: {}".format(st, self.associated.description[:20])
        return "Key: {}".format(st)

    def get_form(self, scope):
        from .forms import KeyForm
        return KeyForm

    def access_view(self, **kwargs):
        kwargs["object"] = self
        kwargs["algo"] = settings.SPIDER_HASH_ALGORITHM.name
        kwargs["hash"] = self.associated.getlist(
            "hash:%s" % settings.SPIDER_HASH_ALGORITHM.name, 1
        )[0]
        return (
            render_to_string(
                "spider_keys/key.html", request=kwargs["request"],
                context=kwargs
            ),
            ""
        )

# Anchors can ONLY be used for server content.
# Clients have to be verified seperately
# 1 To verify client use this trick:
# requesting party redirects user to server, with a return address in GET
# the server generates a secret
# the user authenticates and redirects to the url plus hash of secret
# the server sends secret via post to requesting party
# the requesting party can verify that both: client and server are connected
# this means: anchors are valid for this client
# or 2: generate a token
# or 3: use oauth

# advantage of 1,2: simple
# implementation: if requester in GET, create token and
# redirect after auth to requester with hash of token
# post token to requester

# TODO: make hash algorithm configurable


class AnchorBase(BaseContent):
    expose_name = False
    expose_description = True

    class Meta(BaseContent.Meta):
        abstract = True

    def get_abilities(self, context):
        return {"anchor"}

    def get_priority(self):
        return -10

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret.setdefault("initial", {})
        ret["initial"]["identifier"] = self.get_identifier(kwargs["request"])
        ret["scope"] = kwargs["scope"]
        return ret

    def get_identifier(self, request):
        """ returns id of content, server """
        # security: id can only be faked by own server
        # this should never happen, except with admin access to server
        if not self.associated.id:
            return None
        ret = urljoin(
            "{}://{}".format(request.scheme, get_anchor_domain()), reverse(
                "spider_keys:anchor-permanent",
                kwargs={"pk": self.associated.id}
            )
        )
        return ret


@add_content
class AnchorServer(AnchorBase):
    """ identify by server """
    appearances = [
        {
            "name": "AnchorServer",
            "ctype": VariantType.anchor.value,
            "strength": 0
        }
    ]

    new_url = models.URLField(
        max_length=400, blank=True, null=True,
        help_text=_(
            "Url to new anchor (in case this one is superseded)"
        )
    )
    old_urls = JSONField(
        default=list, blank=True,
        help_text=_(
           "Superseded anchor urls"
        )
    )

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Server-based Anchor")

    def get_size(self):
        s = super().get_size()
        s += 400
        s += len(str(self.old_urls))
        return s

    def get_form(self, scope):
        from .forms import AnchorServerForm
        return AnchorServerForm

    def get_content_name(self):
        return "Anchor: {}".format(self.associated.id)

    def access_anchor(self, **kwargs):
        if self.new_url:
            ret = HttpResponsePermanentRedirect(redirect_to=self.new_url)
        else:
            ret = JsonResponse(self.old_urls)
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


@add_content
class AnchorKey(AnchorBase):
    """ domain name of pc, signed """
    expose_name = False
    expose_description = True
    appearances = [
        {
            "name": "AnchorKey",
            "ctype": VariantType.anchor+VariantType.unique,
            "strength": 0
        }
    ]

    key = models.OneToOneField(
        PublicKey, on_delete=models.CASCADE, related_name="anchorkey",
        help_text=_help_text_key
    )

    signature = models.CharField(
        max_length=1024, help_text=_help_text_sig, null=False
    )

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Key-based Anchor")

    def get_size(self):
        s = super().get_size()
        s += 1024
        return s

    def get_content_name(self):
        st = self.key.get_key_name()
        if st[1]:
            st = st[1]
        else:
            st = "{}...".format(st[0][:10])
        if len(self.key.associated.description) > 0:
            st = "{}: {}".format(st, self.key.associated.description[:20])
        return st

    def get_form(self, scope):
        from .forms import AnchorKeyForm
        return AnchorKeyForm

    def access_anchor(self, **kwargs):
        ret = JsonResponse([])
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def get_info(self):
        ret = super().get_info()
        key = self.key.get_key_name()[0]
        h = get_hashob()
        h.update(key.encode("ascii", "ignore"))
        return "%shash=%s=%s\x1e" % (
            ret, settings.SPIDER_HASH_ALGORITHM.name, h.finalize().hex()
        )


# TODO: implement
# @add_content
class AnchorLink(AnchorBase):
    """
        Anchor by Organisation, e.g. government,
        verifier returns token pointing to url
    """

    class Meta(AnchorBase.Meta):
        abstract = True

    verified_by = models.URLField(max_length=400)

    appearances = [
        {
            "name": "AnchorLink",
            "ctype": VariantType.anchor + VariantType.unique,
            "strength": 10
        }
    ]

    def access_anchor(self, **kwargs):
        ret = JsonResponse(self.verified_by)
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def get_form(self, scope):
        raise NotImplementedError

    def get_info(self):
        return "{}verified_by={}\x1e".format(
            super().get_info(unique=True, unlisted=False),
            self.verified_by
        )



from django.urls import path


from .views import PermAnchorView

app_name = "spider_keys"


urlpatterns = [
    path(
        'anchor/<int:pk>/view/',
        PermAnchorView.as_view(),
        name='anchor-permanent'
    )
]



__all__ = [
    "media_extensions", "image_extensions", "get_anchor_domain",
    "INITIAL_STATIC_TOKEN_SIZE", "STATIC_TOKEN_CHOICES",
    "default_uctoken_duration", "TOKEN_SIZE", "FILE_TOKEN_SIZE",
    "force_captcha", "VALID_INTENTIONS", "VALID_SUB_INTENTIONS"
]

import datetime
from django.conf import settings

from django.utils.translation import gettext_lazy as _

# webbrowser supported image formats
image_extensions = set(getattr(
    settings, "SPIDER_IMAGE_EXTENSIONS", {
        "jpg", "jpeg", "bmp", "png", "ico", "svg", "gif", "webp"
    }
))

media_extensions = set(getattr(
    settings, "SPIDER_MEDIA_EXTENSIONS", {
        "mp4", "ogg", "flac", "mp3", "webm", "wav", "avi"
    }
))

# if None, set to default Site ID if models are ready
_anchor_domain = getattr(
    settings, "SPIDER_ANCHOR_DOMAIN", None
)


def get_anchor_domain():
    global _anchor_domain
    if _anchor_domain:
        return _anchor_domain
    from django.contrib.sites.models import Site
    _anchor_domain = \
        Site.objects.get(id=settings.SITE_ID).domain
    return _anchor_domain


INITIAL_STATIC_TOKEN_SIZE = str(
    getattr(settings, "SPIDER_INITIAL_STATIC_TOKEN_SIZE", 12)
)

STATIC_TOKEN_CHOICES = [
    ("", ""),
    (INITIAL_STATIC_TOKEN_SIZE, _("default ({} Bytes)")),
    ("3", _("low ({} Bytes)")),
    ("12", _("medium ({} Bytes)")),
    ("30", _("high ({} Bytes)")),
]

VALID_INTENTIONS = set(getattr(
    settings, "SPIDER_VALID_INTENTIONS",
    {"auth", "domain", "live", "login", "payment", "persist", "sl"}
))
VALID_SUB_INTENTIONS = {"sl", "live"}

TOKEN_SIZE = getattr(settings, "TOKEN_SIZE", 30)
FILE_TOKEN_SIZE = getattr(settings, "FILE_TOKEN_SIZE", TOKEN_SIZE)

force_captcha = getattr(settings, "REQUIRE_LOGIN_CAPTCHA", False)

# require USE_CAPTCHAS
force_captcha = (force_captcha and getattr(settings, "USE_CAPTCHAS", False))


default_uctoken_duration = getattr(
    settings, "DEFAULT_UCTOKEN_DURATION",
    datetime.timedelta(days=7)
)



__all__ = ["KeyForm", "AnchorServerForm", "AnchorKeyForm"]
# "AnchorGovForm"

import binascii

from django import forms
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext

from cryptography import exceptions
from cryptography.x509 import load_pem_x509_certificate
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.backends import default_backend

from spkcspider.apps.spider.fields import MultipleOpenChoiceField
from spkcspider.apps.spider.widgets import ListWidget

from .models import PublicKey, AnchorServer, AnchorKey


class KeyForm(forms.ModelForm):
    class Meta:
        model = PublicKey
        fields = ['key']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        setattr(self.fields['key'], "hashable", True)

    def clean_key(self):
        _ = gettext
        data = self.cleaned_data['key'].strip()
        if data == "":
            raise forms.ValidationError(
                _('Empty Key'),
                code="empty"
            )
        if "PRIVATE" in data.upper():
            raise forms.ValidationError(
                _('Private Key')
            )
        return data


class AnchorServerForm(forms.ModelForm):
    identifier = forms.CharField(disabled=True)
    anchor_type = forms.CharField(disabled=True, initial="url")
    setattr(identifier, "hashable", True)
    scope = ""
    old_urls = MultipleOpenChoiceField(
        widget=ListWidget(
            format_type="url", item_label=_("Url to superseded anchor")
        ), required=False
    )

    class Meta:
        model = AnchorServer
        fields = ["new_url", "old_urls"]

    def __init__(self, scope, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)
        if self.scope == "add":
            del self.fields["identifier"]
            del self.fields["anchor_type"]
            del self.fields["new_url"]

    def clean_old_urls(self):
        values = self.cleaned_data["old_urls"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format"),
                code='invalid_format'
            )
        return values

    def clean(self):
        _ = gettext
        ret = super().clean()
        if ret.get("new_url", None) and ret.get("old_urls", []):
            raise forms.ValidationError(
                _(
                    "Specify either replacement url or "
                    "urls to superseding anchors"
                ),
                code='invalid_choices'
            )
        return ret


class AnchorKeyForm(forms.ModelForm):
    identifier = forms.CharField(disabled=True)
    anchor_type = forms.CharField(disabled=True, initial="signature")
    scope = ""

    class Meta:
        model = AnchorKey
        fields = ['key', 'signature']

    field_order = ['identifier', 'signature', 'key']

    def __init__(self, scope, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)
        if self.scope == "add":
            del self.fields["identifier"]
            del self.fields["signature"]
            del self.fields["anchor_type"]
            # self.fields["signature"].disabled = True
            # self.fields["signature"].required = False

        if self.scope in ("add", "update"):
            self.fields["key"].queryset = self.fields["key"].queryset.filter(
                models.Q(key__contains="-----BEGIN CERTIFICATE-----") |
                models.Q(key__contains="-----BEGIN PUBLIC KEY-----")
            )
        elif self.scope in ("raw", "list", "view"):
            self.fields["key"] = forms.CharField(
                initial=self.instance.key.key,
                widget=forms.TextArea
            )
            setattr(self.fields['key'], "hashable", True)

    def clean(self):
        _ = gettext
        ret = super().clean()
        try:
            if "-----BEGIN CERTIFICATE-----" in self.cleaned_data["key"].key:
                pubkey = load_pem_x509_certificate(
                    ret["key"].key.encode("utf-8"), default_backend()
                ).public_key()
            else:
                pubkey = serialization.load_pem_public_key(
                    ret["key"].key.encode("utf-8"), default_backend()
                )
        except exceptions.UnsupportedAlgorithm:
            self.add_error("key", forms.ValidationError(
                _("key not usable for signing"),
                code="unusable_key"
            ))
        if self.scope == "add":
            ret["signature"] = "<replaceme>"
            return ret
        chosen_hash = settings.SPIDER_HASH_ALGORITHM
        try:
            pubkey.verify(
                binascii.unhexlify(ret["signature"]),
                ret["identifier"].encode("utf-8"),
                padding.PSS(
                    mgf=padding.MGF1(chosen_hash),
                    salt_length=padding.PSS.MAX_LENGTH
                ),
                chosen_hash
            )
        except exceptions.InvalidSignature:
            self.add_error("signature", forms.ValidationError(
                _("signature incorrect"),
                code="incorrect_signature"
            ))
        except (binascii.Error, KeyError, ValueError):
            self.add_error("signature", forms.ValidationError(
                _("signature malformed or missing"),
                code="malformed_signature"
            ))
        return ret


# class AnchorGovForm(forms.ModelForm):
#    identifier = forms.CharField(disabled=True, widget=forms.HiddenInput())
#    anchor_type = forms.CharField(disabled=True, initial="gov")
#    scope = ""
#
#    class Meta:
#        model = AnchorGov
#        fields = ['idtype', 'token']
#
#    def __init__(self, scope, **kwargs):
#        self.scope = scope
#        super().__init__(**kwargs)
#
#    def clean(self):
#        self.cleaned_data["type"] = "gov"
#        self.cleaned_data["identifier"] = self.instance.get_identifier()
#        self.cleaned_data["format"] = "gov_{verifier}_{token}"
#        if self.scope not in ["add", "update"]:
#            self.cleaned_data["verifier"] = ID_VERIFIERS[self.idtype]
#        del self.cleaned_data["idtype"]
#        return self.cleaned_data



# Generated by Django 2.2 on 2019-04-14 17:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0015_auto_20190414_1743'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assignedcontent',
            name='attached_to_token',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AuthToken'),
        ),
    ]



# Generated by Django 2.2 on 2019-04-25 21:59

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0005_remove_publickey_note'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AnchorKey',
        ),
        migrations.CreateModel(
            name='AnchorKey',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('signature', models.CharField(help_text='Signature of Identifier (hexadecimal-encoded)', max_length=1024)),
                ('key', models.OneToOneField(help_text='"Public Key"-Content for signing identifier. It is recommended to use different keys for signing and encryption.', on_delete=django.db.models.deletion.CASCADE, related_name='anchorkey', to='spider_keys.PublicKey')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='anchorserver',
            name='new_url',
            field=models.URLField(blank=True, help_text='Url to new anchor (in case this one is superseded)', max_length=400, null=True),
        ),
        migrations.AddField(
            model_name='anchorserver',
            name='old_urls',
            field=jsonfield.fields.JSONField(blank=True, default=list, help_text='Superseded anchor urls'),
        ),
    ]



__all__ = ["sitemaps", "ComponentSitemap", "ContentSitemap", "HomeSitemap"]

from django.contrib.sitemaps import GenericSitemap, Sitemap
from django.urls import reverse
from django.conf import settings


class ComponentSitemap(GenericSitemap):
    priority = 0.3
    changefreq = "daily"
    date_field = "modified"
    if not settings.DEBUG:
        protocol = "https"

    def __init__(self):
        from .models import UserComponent
        self.queryset = UserComponent.objects.filter(
            public=True
        )


class ContentSitemap(GenericSitemap):
    priority = 0.7
    changefreq = "hourly"
    date_field = "modified"
    if not settings.DEBUG:
        protocol = "https"

    def __init__(self):
        from .models import AssignedContent
        self.queryset = AssignedContent.objects.filter(
            usercomponent__public=True
        ).exclude(info__contains="_unlisted")


class HomeSitemap(Sitemap):
    priority = 0.5
    changefreq = 'daily'
    if not settings.DEBUG:
        protocol = "https"

    def items(self):
        return ['home']

    def location(self, item):
        return reverse(item)


sitemaps = {
    'components': ComponentSitemap,
    'contents': ContentSitemap,
    'home': HomeSitemap
}



# Generated by Django 2.1.3 on 2018-11-21 07:38

from django.db import migrations


class Migration(migrations.Migration):
    """
        Introduced new field which was not required after integrating the hashes
        in the graph file
    """
    dependencies = [
        ('spider_verifier', '0002_auto_20181031_2121'),
    ]

    operations = [
    ]



__all__ = ["SpiderKeysConfig"]

from django.apps import AppConfig


class SpiderKeysConfig(AppConfig):
    name = 'spkcspider.apps.spider_keys'
    label = 'spider_keys'
    spider_url_path = 'spiderkeys/'
    verbose_name = 'spkcspider keys, identifiers and anchors'



# Generated by Django 2.2.1 on 2019-05-12 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0021_auto_20190511_1046'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authtoken',
            name='created_by_special_user',
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='login_protection',
            field=models.CharField(choices=[('a', 'Hide'), ('b', 'Hide if triggered'), ('c', 'Disable login'), ('d', 'Wipe'), ('e', 'Wipe User')], default='a', max_length=1),
        ),
    ]



__all__ = ["PermAnchorView"]

from urllib.parse import urljoin

from django.http.response import (
    HttpResponsePermanentRedirect, HttpResponseBase
)
from django.views.generic.detail import DetailView

from spkcspider.apps.spider.conf import get_anchor_domain

from spkcspider.apps.spider.models import AssignedContent


class PermAnchorView(DetailView):
    queryset = AssignedContent.objects.filter(
        info__contains="\x1eanchor\x1e"
    )

    def get(self, request, *args, **kwargs):
        if request.get_host() != get_anchor_domain():
            return HttpResponsePermanentRedirect(
                redirect_to=urljoin(
                    get_anchor_domain(), request.get_full_path()
                )
            )
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs["scope"] = "anchor"
        return super().get_context_data(**kwargs)

    def render_to_response(self, context):
        ret = self.object.content.access(context)
        assert(isinstance(ret, HttpResponseBase))
        return ret



# Generated by Django 2.1.7 on 2019-02-25 20:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0007_usercomponent_allow_domain_mode'),
    ]

    def remove_breaking(apps, schema_editor):
        AuthToken = apps.get_model('spider_base', 'AuthToken')
        AuthToken.objects.filter(referrer__isnull=False).delete()

    operations = [
        migrations.CreateModel(
            name='ReferrerObject',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('url', models.URLField(db_index=True, editable=False, max_length=600, unique=True)),
            ],
        ),
        migrations.RunPython(remove_breaking),
        migrations.AlterField(
            model_name='authtoken',
            name='referrer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tokens', to='spider_base.ReferrerObject'),
        ),
    ]



"""
installed_protections, add_protection
namespace: spider_base

"""

__all__ = ("installed_protections", "BaseProtection", "ProtectionResult",
           "initialize_protection_models")

import logging
import binascii
from random import SystemRandom
from hashlib import sha256
from base64 import b64encode, b64decode
import ipaddress

from django.conf import settings
from django import forms

from django.apps import apps as django_apps
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from django.contrib.auth import authenticate
from django.utils.crypto import constant_time_compare
from django.views.decorators.debug import sensitive_variables
# from django.contrib.auth.hashers import make_password

import ratelimit
from ratelimit import parse_rate
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

from .helpers import add_by_field, create_b64_token, aesgcm_pbkdf2_cryptor
from .constants import ProtectionType, ProtectionResult
from .fields import MultipleOpenChoiceField
from .widgets import OpenChoiceWidget, PWOpenChoiceWidget

logger = logging.getLogger(__name__)
installed_protections = {}
_sysrand = SystemRandom()

# don't spam set objects
_empty_set = frozenset()

# for debug/min switch
_extra = '' if settings.DEBUG else '.min'

_pbkdf2_params = {
    "iterations": 120000,
    "hash_name": "sha512",
    "dklen": 32
}

_Scrypt_params = {
    "length": 64,
    "n": 2**14,
    "r": 16,
    "p": 2,
}


def initialize_protection_models(apps=None):
    if not apps:
        apps = django_apps
    UserComponent = apps.get_model("spider_base", "UserComponent")
    AssignedProtection = apps.get_model("spider_base", "AssignedProtection")
    ProtectionModel = apps.get_model("spider_base", "Protection")
    for code, val in installed_protections.items():
        ret = ProtectionModel.objects.get_or_create(
            defaults={"ptype": val.ptype}, code=code
        )[0]
        if ret.ptype != val.ptype:
            ret.ptype = val.ptype
            ret.save()

    login = ProtectionModel.objects.filter(code="login").first()
    if login:
        for uc in UserComponent.objects.filter(name="index"):
            asuc = AssignedProtection.objects.get_or_create(
                defaults={"active": True},
                usercomponent=uc, protection=login
            )[0]
            if not asuc.active:
                asuc.active = True
                asuc.save()

    invalid_models = ProtectionModel.objects.exclude(
        code__in=installed_protections.keys()
    )
    if invalid_models.exists():
        print("Invalid protections, please update or remove them:",
              [t.code for t in invalid_models])


class BaseProtection(forms.Form):
    """
        Base for Protections
        Usage: use form to define some configurable fields for configuration
        use auth to validate: in this case:
            template_name, render, form variable are used
    """
    use_required_attribute = False
    active = forms.BooleanField(label=_("Active"), required=False)
    instant_fail = forms.BooleanField(
        label=_("Instant fail"), required=False,
        help_text=_("Fail instantly if not fullfilled. "
                    "Don't count to required_passes.")
    )
    # unique code name max 10 slug chars
    # if imported by extract_app_dicts, name is automatically set to key name
    # name = foo

    # ptype valid for, is overwritten with current ptype
    ptype = ProtectionType.access_control.value

    # description of Protection
    description = None

    template_name = None
    # form for authentication
    form = None
    # optional render function
    # render = None

    # auto populated, instance
    protection = None
    instance = None
    usercomponent = None
    parent_form = None

    # initial values
    initial = {}

    def __init__(self, *, protection, ptype, request, uc, form, **kwargs):
        AssignedProtection = django_apps.get_model(
            "spider_base", "AssignedProtection"
        )
        self.ptype = ptype
        self.parent_form = form
        self.instance = uc.protections.filter(
            protection=protection
        ).first()
        if not self.instance:
            self.instance = AssignedProtection(
                usercomponent=uc, protection=protection, active=False
            )
        initial = self.get_initial()
        # use instance informations
        initial["active"] = self.instance.active
        initial["instant_fail"] = self.instance.instant_fail
        # copy of initial is saved as self.initial, so safe to change it
        # after __init__ is called
        super().__init__(initial=initial, **kwargs)
        self.fields["active"].help_text = self.description

    def get_initial(self):
        initial = self.initial.copy()
        initial.update(self.instance.data)
        return initial

    def get_strength(self):
        # can provide strength in range 0-4
        # 0 no protection
        # 1 weak protection
        # 2 normal protection
        # 3 strong protection
        # 4 reserved for login only, can be returned to auth user
        # tuple for min max
        return (1, 1)

    @staticmethod
    def extract_form_kwargs(request):
        kwargs = {}
        if request.method in ["POST", "PUT"]:
            kwargs = {
                'data': request.POST,
                'files': request.FILES,
            }
        return kwargs

    @classmethod
    def auth(cls, request, **kwargs):
        if hasattr(cls, "auth_form"):
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if form.is_valid():
                return 1
            return form
        return False

    @classmethod
    def localize_name(cls, name=None):
        if not name:
            name = cls.name
        return pgettext("protection name", name.title())

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name(name)

    def __str__(self):
        return self.localize_name(self.name)

    def save(self):
        self.instance.active = self.cleaned_data.pop("active")
        self.instance.instant_fail = self.cleaned_data.pop("instant_fail")
        self.instance.data = self.cleaned_data
        self.instance.save()


# only friends have access
@add_by_field(installed_protections, "name")
class FriendProtection(BaseProtection):
    name = "friends"
    ptype = ProtectionType.access_control.value

    users = MultipleOpenChoiceField(
        label=_("Users"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    description = _("Limit access to selected users")

    def get_strength(self):
        return (3, 3)

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if (
            obj and request.user.is_authenticated and
            (
                getattr(request.user, request.user.USERNAME_FIELD) in
                obj.data["users"]
            )
        ):
            return 3
        else:
            return False


@add_by_field(installed_protections, "name")
class RandomFailProtection(BaseProtection):
    name = "randomfail"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value
    ptype += ProtectionType.side_effects.value

    success_rate = forms.IntegerField(
        label=_("Success Rate"), min_value=20, max_value=100, initial=100,
        widget=forms.NumberInput(attrs={'type': 'range'}),
        help_text=_("Set success rate")
    )

    use_404 = forms.BooleanField(label="Use 404 errors?", required=False)

    description = _(
        "Fail/Refuse randomly. Optionally with 404 error, "
        "to disguise correct access."
    )

    def get_strength(self):
        return (0, 0)

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Random Fail")

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if obj and obj.data.get("success_rate", None):
            if _sysrand.randrange(1, 101) <= obj.data["success_rate"]:
                return 0
            elif obj.data.get("use_404", False):
                raise Http404()
        return False


@add_by_field(installed_protections, "name")
class RateLimitProtection(BaseProtection):
    name = "ratelimit"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value

    rate_accessed = forms.RegexField(
        regex=r'([\d]+)/([\d]*)([smhd])?',
        required=False,
        help_text=_(
            "Maximal access tries to this component per ip "
            "before blocking. Format: tries/multiplier"
            "(smhdw, second till week)"
        )
    )
    rate_static_token_error = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal http404 errors per hour per user/ip before blocking"
        )
    )
    rate_login_failed_ip = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal failed logins per ip per hour before blocking"
        )
    )
    rate_login_failed_account = forms.IntegerField(
        required=False, min_value=1,
        help_text=_(
            "Maximal failed logins per hour for this user before blocking. "
            "Useful for login"
        )
    )

    description = _(
        "Ratelimit access tries"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if ProtectionType.authentication.value in self.ptype:
            pass

    def get_strength(self):
        return (1, 1)

    def clean_rate_accessed(self):
        try:
            ret = parse_rate(self.cleaned_data.get("rate_accessed"))
            if ret[0] <= 0 or ret[1] <= 0:
                return None
        except ValueError:
            return None
        return self.cleaned_data.get("rate_accessed")

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Rate Limit")

    @classmethod
    def count_access(cls, request, obj):
        return b"%b:%b" % (
            str(obj.usercomponent.pk).encode("ascii"), ipaddress.ip_network(
                request.META['REMOTE_ADDR'], strict=False
            ).compressed.encode("ascii")
        )

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return False
        if obj:
            temp = obj.data.get("rate_accessed", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_ratelimit_accessed",
                key=cls.count_access(request, obj), rate=temp, inc=True
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_static_token_error", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_static_token_error",
                key="user_or_ip", rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_login_failed_ip", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_login_failed_ip",
                key="ip", rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
            temp = obj.data.get("rate_login_failed_account", None)
            if temp and ratelimit.get_ratelimit(
                request=request, group="spider_login_failed_account",
                key=lambda x, y: obj.usercomponent.username,
                rate=(int(temp), 3600), inc=False
            )["request_limit"] > 0:
                return False
        return 1


@add_by_field(installed_protections, "name")
class LoginProtection(BaseProtection):
    name = "login"
    ptype = ProtectionType.authentication.value
    ptype += ProtectionType.access_control.value

    description = _("Use Login password")

    allow_auth = forms.BooleanField(
        label=_("Component authentication"), required=False
    )
    _request = None

    class auth_form(forms.Form):
        use_required_attribute = False
        password = forms.CharField(
            label=_("Password"),
            strip=False,
            widget=forms.PasswordInput,
        )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._request = kwargs.get("request")
        if ProtectionType.authentication.value in self.ptype:
            self.fields["active"].initial = True
            self.initial["active"] = True
            self.fields["active"].disabled = True
            del self.fields["allow_auth"]

    def clean(self):
        super().clean()
        if ProtectionType.authentication.value in self.ptype:
            if not authenticate(
                self._request, username=self.instance.usercomponent.username,
                password=self.parent_form.initial["master_pw"], nospider=True
            ):
                self.parent_form.add_error(
                    "master_pw",
                    forms.ValidationError(
                        _("Invalid Password"),
                        code="invalid_password"
                    )
                )
        return self.cleaned_data

    def get_strength(self):
        return (3, 4 if self.cleaned_data.get("allow_auth", False) else 3)

    @classmethod
    @sensitive_variables("password")
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return cls.auth_form()

        username = obj.usercomponent.username
        for password in request.POST.getlist("password")[:4]:
            if authenticate(
                request, username=username, password=password, nospider=True
            ):
                if obj.data.get("allow_auth", False):
                    return 4
                return 3
        return cls.auth_form()

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name("Password")


@add_by_field(installed_protections, "name")
class PasswordProtection(BaseProtection):
    name = "password"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value

    description = _("Protect with extra passwords")
    prefix = "protection_passwords"

    class Media:
        css = {
            'all': [
                'node_modules/select2/dist/css/select2.min.css'
            ]
        }
        js = [
            'node_modules/base64-js/base64js.min.js',
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            'node_modules/select2/dist/js/select2%s.js' % _extra,
            'spider_base/protections/PasswordProtection.js'
        ]

    class auth_form(forms.Form):
        use_required_attribute = False
        password = forms.CharField(
            label=_("Extra Password"),
            strip=False,
            widget=forms.PasswordInput,
        )

    salt = forms.CharField(
        widget=forms.HiddenInput
    )
    default_master_pw = forms.CharField(
        widget=forms.HiddenInput, disabled=True
    )

    passwords = MultipleOpenChoiceField(
        label=_("Passwords"), required=False,
        widget=PWOpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    auth_passwords = MultipleOpenChoiceField(
        label=_("Passwords (for component authentication)"), required=False,
        widget=PWOpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    _successful_clean = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        salt = self.initial.get("salt")
        if not salt:
            salt = self.fields["salt"].widget.value_from_datadict(
                self.data, self.files, self.add_prefix("salt")
            )
        if not salt:
            salt = create_b64_token()
        self.initial["salt"] = salt
        self.initial["default_master_pw"] = sha256(
            "".join([salt, settings.SECRET_KEY]).encode("utf-8")
        ).hexdigest()
        if ProtectionType.authentication.value in self.ptype:
            del self.fields["auth_passwords"]

    @staticmethod
    def eval_strength(length):
        if not length:
            return 0
        elif length > 15:
            return 2
        elif length > 40:
            return 3
        return 1

    def get_strength(self):
        maxstrength = self.eval_strength(self.cleaned_data["max_length"])
        if len(self.cleaned_data["auth_passwords"]) > 0:
            maxstrength = 4
        return (
            self.eval_strength(self.cleaned_data["min_length"]),
            maxstrength
        )

    @classmethod
    def hash_pw(cls, pw, salt, params=_Scrypt_params):
        return b64encode(Scrypt(
            salt=salt,
            backend=default_backend(),
            **params
        ).derive(pw[:128].encode("utf-8"))).decode("ascii")

    def is_valid(self):
        if not self._successful_clean:
            return False
        return super().is_valid()

    def clean(self):
        super().clean()
        # prevents user self lockout
        if ProtectionType.authentication.value in self.ptype and \
           len(self.cleaned_data["passwords"]) == 0:
            self.cleaned_data["active"] = False
        elif (
            len(self.cleaned_data["passwords"]) == 0 and
            len(self.cleaned_data["auth_passwords"]) == 0
        ):
            self.cleaned_data["active"] = False

        if getattr(self.instance, "id", None) and not self.has_changed():
            self.cleaned_data.update(self.instance.data)
            self._successful_clean = True
            return self.cleaned_data

        # eliminate duplicates
        self.cleaned_data["passwords"] = list(set(
            self.cleaned_data.get("passwords", _empty_set)
        ))
        self.cleaned_data["auth_passwords"] = list(set(
            self.cleaned_data.get("auth_passwords", _empty_set)
        ))
        self.cleaned_data["pbkdf2_params"] = _pbkdf2_params

        min_length = None
        max_length = None
        cryptor = None
        salt = self.cleaned_data.get("salt", "").encode("ascii")
        if not salt:
            return self.cleaned_data
        master_pw = self.parent_form.initial.get("master_pw", None)
        self.cleaned_data.pop("default_master_pw", None)
        has_master_pw = True
        if not master_pw:
            has_master_pw = False
            master_pw = self.initial["default_master_pw"]
        cryptor = aesgcm_pbkdf2_cryptor(
            master_pw, salt=salt,
            params=self.cleaned_data["pbkdf2_params"]
        )

        hashed_passwords = []
        auth_passwords = []
        hashed_auth_passwords = []
        current_field = "passwords"
        try:
            for pw in self.cleaned_data["passwords"]:
                if pw.startswith("bogo"):
                    raise ValueError("bogo not allowed here")
                nonce, pw = map(b64decode, pw.split(":", 1))
                pw = cryptor.decrypt(nonce, pw, None).decode("utf-8")
                lenpw = len(pw)
                if not min_length or lenpw < min_length:
                    min_length = lenpw
                if not max_length or lenpw > max_length:
                    max_length = lenpw
                hashed_passwords.append(self.hash_pw(
                    pw, salt, params=_Scrypt_params
                ))

            current_field = "auth_passwords"

            for pw in self.cleaned_data["auth_passwords"]:
                if pw.startswith("bogo"):
                    raise ValueError("bogo not allowed here")
                pwsource = pw
                nonce, pw = map(b64decode, pw.split(":", 1))
                pw = cryptor.decrypt(nonce, pw, None).decode("utf-8")
                lenpw = len(pw)
                if self.eval_strength(lenpw) < 2:
                    continue
                if not min_length or lenpw < min_length:
                    min_length = lenpw
                if not max_length or lenpw > max_length:
                    max_length = lenpw
                auth_passwords.append(pwsource)
                hashed_auth_passwords.append(self.hash_pw(
                    pw, salt, params=_Scrypt_params
                ))

            self.cleaned_data["hashed_passwords"] = hashed_passwords
            self.cleaned_data["auth_passwords"] = auth_passwords
            self.cleaned_data["scrypt_params"] = _Scrypt_params
            self.cleaned_data["hashed_auth_passwords"] = hashed_auth_passwords
            self.cleaned_data["min_length"] = min_length
            self.cleaned_data["max_length"] = max_length
            self._successful_clean = True
        except InvalidTag:
            if has_master_pw:
                self.parent_form.add_error(
                    "master_pw",
                    forms.ValidationError(
                        _(
                            "Invalid master password? "
                            "Passwords were not readable"
                        )
                    )
                )
            else:
                self.add_error(
                    current_field,
                    forms.ValidationError(
                        _(
                            "Forgot to enter master password? "
                            "Passwords were not readable"
                        )
                    )
                )
        except (ValueError, binascii.Error) as exc:
            logger.warning("PWProtection: pw encoding problems", exc_info=exc)
            self.add_error(
                current_field,
                forms.ValidationError(
                    _("Format error, check passwords")
                )
            )

        return self.cleaned_data

    @classmethod
    @sensitive_variables("password", "pw")
    def auth(cls, request, obj, **kwargs):
        retfalse = cls.auth_form()
        if not obj:
            retfalse.fields["password"].label = _(
                "Extra Password (if required)"
            )
            return retfalse
        success = False
        auth = False
        max_length = 0
        salt = obj.data.get("salt", "").encode("ascii")
        for password in request.POST.getlist("password")[:4]:
            if salt:
                pwhash = cls.hash_pw(
                    password, salt, params=obj.data.get(
                        "scrypt_params", _Scrypt_params
                    )
                )

                for pw in obj.data["hashed_passwords"]:
                    if constant_time_compare(pw, pwhash):
                        success = True

                for pw in obj.data["hashed_auth_passwords"]:
                    if constant_time_compare(pw, pwhash):
                        success = True
                        auth = True
            else:

                for pw in obj.data["passwords"]:
                    if constant_time_compare(pw, password):
                        success = True

                for pw in obj.data.get("auth_passwords", []):
                    if constant_time_compare(pw, password):
                        success = True
                        auth = True
            if success:
                max_length = max(len(password), max_length)

        if success:
            if auth:
                return 4
            return cls.eval_strength(max_length)
        return retfalse


if getattr(settings, "USE_CAPTCHAS", False):
    from captcha.fields import CaptchaField

    @add_by_field(installed_protections, "name")
    class CaptchaProtection(BaseProtection):
        name = "captcha"
        ptype = ProtectionType.access_control.value
        ptype += ProtectionType.authentication.value
        description = _("Require captcha")

        class auth_form(forms.Form):
            use_required_attribute = False
            prefix = "protection_captcha"

        auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            CaptchaField(label=_("Captcha"))
        auth_form.base_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME]

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            if ProtectionType.authentication.value in self.ptype:
                # login should not be possible with captcha alone
                self.initial["instant_fail"] = True
                self.fields["instant_fail"].initial = True
                self.fields["instant_fail"].disabled = True
                self.fields["instant_fail"].help_text = \
                    _("instant_fail is for login required")

                if getattr(settings, "REQUIRE_LOGIN_CAPTCHA", False):
                    self.initial["active"] = True
                    self.fields["active"].disabled = True
                    self.fields["instant_fail"].help_text = \
                        _("captcha is for login required (admin setting)")

        def get_strength(self):
            return (1, 1)

        @classmethod
        def localize_name(cls, name=None):
            return pgettext("protection name", "Captcha Protection")

        @classmethod
        def auth(cls, request, obj, **kwargs):
            if not obj:
                return cls.auth_form()
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if request.method != "GET" and form.is_valid():
                return 1
            return form



# Generated by Django 2.1.7 on 2019-03-09 22:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0005_auto_20190302_1741'),
    ]

    operations = [
        migrations.AddField(
            model_name='verifysourceobject',
            name='update_secret',
            field=models.CharField(blank=True, max_length=120, null=True, unique=True),
        ),
    ]



__all__ = (
    "DEFAULT_LICENSE_FILE", "DEFAULT_LICENSE_TEXT", "LICENSE_CHOICES"
)

from django.conf import settings

LICENSE_CHOICES = getattr(
    settings, "SPIDER_LICENSE_CHOICES", {}
)
DEFAULT_LICENSE_FILE = getattr(
    settings, "SPIDER_DEFAULT_LICENSE_FILE",
    lambda uc, user: "other"
)
if not callable(DEFAULT_LICENSE_FILE):
    _DEFAULT_LICENSE_FILE = DEFAULT_LICENSE_FILE

    def DEFAULT_LICENSE_FILE(uc, user):
        return _DEFAULT_LICENSE_FILE

DEFAULT_LICENSE_TEXT = getattr(
    settings, "SPIDER_DEFAULT_LICENSE_TEXT", DEFAULT_LICENSE_FILE
)

if not callable(DEFAULT_LICENSE_TEXT):
    _DEFAULT_LICENSE_TEXT = DEFAULT_LICENSE_TEXT

    def DEFAULT_LICENSE_TEXT(uc, user):
        return _DEFAULT_LICENSE_TEXT



"""
Basic Contents like TravelProtection, or Links
namespace: spider_base

"""

__all__ = [
    "LinkContent", "PersistenceFeature", "TravelProtection"
]

import logging
from datetime import timedelta
from base64 import b64encode

from django.conf import settings
from django.db import models
from django.utils.html import escape
from django.shortcuts import redirect
from django.utils.translation import gettext
from django.middleware.csrf import CsrfViewMiddleware
from django.http.response import HttpResponseBase
from django.utils import timezone
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext


from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

from ..contents import BaseContent, add_content
from ..constants import (
    TravelLoginType, VariantType, ActionUrl, travel_scrypt_params,
    dangerous_login_choices
)

logger = logging.getLogger(__name__)

# raw and export: use references
link_abilities = frozenset(["add", "update", "update_raw", "raw", "export"])


@add_content
class PersistenceFeature(BaseContent):
    appearances = [
        {
            "name": "Persistence",
            "ctype": VariantType.component_feature.value,
            "strength": 0
        },
    ]

    class Meta:
        abstract = True

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl(
                "renew-token",
                reverse("spider_base:token-renew")
            ),
            ActionUrl(
                "delete-token",
                reverse("spider_base:token-delete-request")
            )
        ]


@add_content
class DomainModeFeature(BaseContent):
    appearances = [
        {
            "name": "DomainMode",
            "ctype": (
                VariantType.component_feature + VariantType.content_feature +
                VariantType.no_export
            ),
            "valid_feature_for": "*",
            "strength": 0
        },
    ]

    class Meta:
        abstract = True

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl(
                "delete-token",
                reverse("spider_base:token-delete-request")
            )
        ]


@add_content
class LinkContent(BaseContent):
    appearances = [{
        "name": "Link",
        "ctype": VariantType.raw_update.value
    }]
    expose_name = False
    expose_description = False

    content = models.ForeignKey(
        "spider_base.AssignedContent", related_name="+",
        on_delete=models.CASCADE,
        limit_choices_to={
            "strength_link__lte": 10,
        }
    )

    push = models.BooleanField(
        blank=True, default=False,
        help_text=_("Improve ranking of this Link.")
    )

    def get_abilities(self, context):
        ret = set()
        if self.associated.getflag("anchor"):
            ret.add("anchor")
        return ret

    def get_content_name(self):
        return self.content.content.get_content_name()

    def get_content_description(self):
        return self.content.content.get_content_description()

    def get_strength(self):
        return self.content.content.get_strength()

    def get_priority(self):
        priority = self.content.content.get_priority()
        # pin to top
        if self.push and priority < 1:
            return 1
        return priority

    def get_strength_link(self):
        # don't allow links linking on links
        return 11

    def get_info(self):
        ret = self.content.content.get_info()
        return "%ssource=%s\x1elink\x1e" % (
            ret, self.associated.pk
        )

    def get_references(self):
        if not self.content:
            return []
        return [self.content]

    def get_form(self, scope):
        from ..forms import LinkForm
        if scope in link_abilities:
            return LinkForm
        # maybe not required
        return self.content.content.get_form(scope)

    def get_form_kwargs(self, **kwargs):
        if kwargs["scope"] in link_abilities:
            ret = super().get_form_kwargs(**kwargs)
            ret["uc"] = kwargs["uc"]
            ret["request"] = kwargs["request"]
        else:
            # maybe not required anymore
            ret = self.content.content.get_form_kwargs(**kwargs)
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Content Link"))
        return super().access_add(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Content Link"))
        return super().access_update(**kwargs)

    def access_raw_update(self, **kwargs):
        return redirect(
            'spider_base:ucontent-access',
            token=self.content.token,
            access='update'
        )

    def access(self, context):
        # context is updated and used outside!!
        # so make sure that func gets only a copy (**)
        context.setdefault("extra_outer_forms", [])
        func = self.access_default
        if context["scope"] in link_abilities:
            func = getattr(self, "access_{}".format(context["scope"]))

            # check csrf tokens manually
            csrferror = CsrfViewMiddleware().process_view(
                context["request"], None, (), {}
            )
            if csrferror is not None:
                # csrferror is HttpResponse
                return csrferror
            ret = func(**context)
            if context["scope"] == "update":
                # update function should never return HttpResponse
                #  (except csrf error)
                assert(not isinstance(ret, HttpResponseBase))
            return ret
        else:
            context["source"] = self
            context["uc"] = self.content.usercomponent
            ret = self.access(context)
            context["uc"] = self.usercomponent
            return ret


login_choices = [
    (TravelLoginType.hide.value, _("Hide")),
    (TravelLoginType.trigger_hide.value, _("Hide if triggered")),
    (TravelLoginType.disable.value, _("Disable login")),
    (TravelLoginType.trigger_disable.value, _("Disable login if triggered")),
    (TravelLoginType.wipe.value, _("Wipe")),
    (TravelLoginType.wipe_user.value, _("Wipe User")),
]

login_choices_dict = dict(login_choices)


class TravelProtectionManager(models.Manager):
    def get_active(self, now=None):
        if not now:
            now = timezone.now()
        q = models.Q(active=True)
        q &= (
            models.Q(start__isnull=True) | models.Q(start__lte=now)
        )
        q &= (models.Q(stop__isnull=True) | models.Q(stop__gte=now))

        q &= (
            models.Q(approved=True) |
            ~models.Q(login_protection__in=dangerous_login_choices)
        )
        return self.filter(q)

    def get_active_for_session(self, session, user, now=None):
        q = ~models.Q(associated_rel__info__contains="\x1epwhash=")
        if user.is_authenticated:
            for pw in session.get("travel_hashed_pws", []):
                q |= models.Q(
                    associated_rel__info__contains="\x1epwhash=%s\x1e" % pw
                )
            q &= models.Q(associated_rel__usercomponent__user=user)
        return self.get_active(now).filter(q)

    def get_active_for_request(self, request, now=None):
        return self.get_active_for_session(request.session, request.user)

    def auth(self, request, uc, now=None):
        active = self.get_active(now).filter(
            associated_rel__usercomponent=uc
        )

        request.session["travel_hashed_pws"] = list(map(
            lambda x: b64encode(Scrypt(
                salt=settings.SECRET_KEY.encode("utf-8"),
                backend=default_backend(),
                **travel_scrypt_params
            ).derive(x[:128].encode("utf-8"))).decode("ascii"),
            request.POST.getlist("password")[:4]
        ))

        q = ~models.Q(associated_rel__info__contains="\x1epwhash=")
        for pw in request.session["travel_hashed_pws"]:
            q |= models.Q(
                associated_rel__info__contains="\x1epwhash=%s\x1e" % pw
            )

        active = active.filter(q)

        request.session["is_travel_protected"] = active.exists()

        # use cached result instead querying
        if not request.session["is_travel_protected"]:
            return True

        for i in active:
            if TravelLoginType.disable.value == i.login_protection:
                return False
            elif TravelLoginType.trigger_hide.value == i.login_protection:
                i.login_protection = TravelLoginType.hide.value
                # don't re-add trigger passwords here
                if i.associated.getflag("anonymous_deactivation"):
                    i._encoded_form_info = \
                        "{}anonymous_deactivation\x1e".format(
                            i._encoded_form_info
                        )
                if i.associated.getflag("anonymous_trigger"):
                    i._encoded_form_info = \
                        "{}anonymous_trigger\x1e".format(
                            i._encoded_form_info
                        )
                i.clean()
                # assignedcontent is fully updated
                i.save(update_fields=["login_protection"])
            elif TravelLoginType.trigger_disable.value == i.login_protection:
                i.login_protection = TravelLoginType.disable.value
                # don't re-add trigger passwords here
                if i.associated.getflag("anonymous_deactivation"):
                    i._encoded_form_info = \
                        "{}anonymous_deactivation\x1e".format(
                            i._encoded_form_info
                        )
                if i.associated.getflag("anonymous_trigger"):
                    i._encoded_form_info = \
                        "{}anonymous_trigger\x1e".format(
                            i._encoded_form_info
                        )
                i.clean()
                # assignedcontent is fully updated
                i.save(update_fields=["login_protection"])
            elif TravelLoginType.wipe_user.value == i.login_protection:
                uc.user.delete()
                return False
            elif TravelLoginType.wipe.value == i.login_protection:
                # first components have to be deleted
                i.protect_components.all().delete()
                # as this deletes itself and therefor the information
                # about affected components
                i.protect_contents.all().delete()
        return True


@add_content
class TravelProtection(BaseContent):
    # should not be a feature as it would be detectable this way
    appearances = [
        {
            "name": "TravelProtection",
            "strength": 10,
        },
        {
            "name": "SelfProtection",
            "strength": 10,
        }
    ]

    objects = TravelProtectionManager()

    active = models.BooleanField(default=False, blank=True)
    # no start for always valid = self protection
    start = models.DateTimeField(blank=True, null=True)
    # no stop for no termination
    stop = models.DateTimeField(blank=True, null=True)
    approved = models.BooleanField(default=False, blank=True)

    login_protection = models.CharField(
        max_length=1, choices=login_choices,
        default=TravelLoginType.hide.value
    )

    protect_components = models.ManyToManyField(
        "spider_base.UserComponent", related_name="travel_protected",
        blank=True, limit_choices_to={
            "strength__lt": 10,
        }
    )
    protect_contents = models.ManyToManyField(
        "spider_base.AssignedContent", related_name="travel_protected",
        blank=True
    )

    force_token_size = 60
    expose_name = False
    expose_description = False

    _anonymous_deactivation = False
    _encoded_form_info = ""

    class Meta:
        default_permissions = ()
        permissions = [
            (
                "approve_travelprotection",
                "Can approve dangerous TravelProtections"
            )
        ]

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        if name == "TravelProtection":
            return _("content name", "Travel-Protection")
        else:
            return _("content name", "Self-Protection")

    def get_content_name(self):
        if self.start and self.stop:
            if self.stop - self.start < timedelta(days=1):
                return "{}:{}–{}".format(
                    login_choices_dict[self.login_protection],
                    self.start.time(), self.stop.time()
                )
            return "{}:{}–{}".format(
                login_choices_dict[self.login_protection],
                self.start.date(), self.stop.date()
            )
        elif self.start:
            return "{}:{}–{}".format(
                login_choices_dict[self.login_protection],
                self.start.date()
            )
        else:
            return "{}: -".format(
                login_choices_dict[self.login_protection]
            )

        stop = self.stop or "-"
        return "{}:{}–{}".format(
            login_choices_dict[self.login_protection],
            self.start.date(), stop != "-" and stop.date()
        )

    def get_strength_link(self):
        return 11

    def get_priority(self):
        # pin to top with higher priority
        return 2

    def get_form(self, scope):
        from ..forms import TravelProtectionForm
        return TravelProtectionForm

    def get_form_kwargs(self, **kwargs):
        ret = super().get_form_kwargs(**kwargs)
        ret["request"] = kwargs["request"]
        return ret

    def access_add(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Create Travel Protection"))
        # token = self.associated.token
        ret = super().access_add(**kwargs)
        # if (
        #     token != self.associated.token and
        #     TravelProtection.objects.get_active().filter(pk=self.pk).exists()
        # ):
        #     return redirect(
        #         'spider_base:ucontent-list', kwargs={
        #            "token": kwargs.get("source", self.usercomponent).token
        #        }
        #    )
        return ret

    def access_update(self, **kwargs):
        _ = gettext
        kwargs["legend"] = escape(_("Update Travel Protection"))
        # token = self.associated.token
        ret = super().access_update(**kwargs)
        # if (
        #     token != self.associated.token and
        #     TravelProtection.objects.get_active().filter(pk=self.pk).exists()
        # ):
        #     return redirect(
        #         'spider_base:ucontent-list', kwargs={
        #             "token": kwargs.get("source", self.usercomponent).token
        #        }
        #    )
        return ret

    def get_info(self):
        return "{}{}".format(
            super().get_info(unlisted=True),
            self._encoded_form_info
        )



# Generated by Django 2.2 on 2019-04-14 14:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0006_auto_20190226_1853'),
    ]

    operations = [
        migrations.AddField(
            model_name='taglayout',
            name='description',
            field=models.TextField(blank=True, default=''),
        ),
    ]



__all__ = ["SchemeWidget"]

import json

from django.forms import widgets
from django.conf import settings


_extra = '' if settings.DEBUG else '.min'


class SchemeWidget(widgets.Textarea):
    template_name = 'spider_base/forms/widgets/wrapped_textarea.html'

    class Media:
        js = [
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            'node_modules/select2/dist/js/select2%s.js' % _extra,
            'node_modules/@json-editor/json-editor/dist/jsoneditor%s.js' % _extra,  # noqa:E501,
            'spider_tags/scheme_editor.js'
        ]
        css = {
            'all': [
                'node_modules/select2/dist/css/select2%s.css' % _extra
            ]
        }

    def __init__(self, *, attrs=None, wrapper_attrs=None, **kwargs):
        if not attrs:
            attrs = {"class": ""}
        if not wrapper_attrs:
            wrapper_attrs = {}
        attrs.setdefault("class", "")
        attrs["class"] += " SchemeEditorTarget"
        self.wrapper_attrs = wrapper_attrs.copy()
        super().__init__(attrs=attrs, **kwargs)

    def __deepcopy__(self, memo):
        obj = super().__deepcopy__(memo)
        obj.wrapper_attrs = self.wrapper_attrs.copy()
        return obj

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget']['wrapper_attrs'] = self.wrapper_attrs
        context['widget']['wrapper_attrs']["id"] = "{}_inner_wrapper".format(
            context['widget']['attrs']["id"]
        )
        return context

    def format_value(self, value):
        if not value:
            return "[]"
        if isinstance(value, (tuple, list)):
            value = json.dumps(value)
        return str(value)

    def render(self, name, value, attrs=None, renderer=None):
        if value is None:
            value = ""
        if not isinstance(value, str):
            value = json.dumps(value, ensure_ascii=False, indent=2)

        return super().render(name, value, attrs, renderer)



# Generated by Django 2.1.2 on 2018-11-11 13:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0003_dataverificationtag_linked_hashes'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='dataverificationtag',
            name='linked_hashes',
        ),
    ]



"""
User Components/Info
namespace: spider_base

"""

__all__ = [
    "UserComponent", "UserComponentManager", "TokenCreationError",
    "UserInfo"
]

import logging

from django.db import models
from django.conf import settings
from django.apps import apps
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from django.urls import reverse
from django.core.exceptions import ValidationError
from django.core import validators

from ..helpers import get_settings_func, create_b64_id_token
from ..validators import validator_token
from ..constants import (
    ProtectionType, VariantType, MAX_TOKEN_B64_SIZE, TokenCreationError,
    hex_size_of_bigid, static_token_matcher
)
from ..conf import default_uctoken_duration, force_captcha


logger = logging.getLogger(__name__)


_name_help = _(
    "Name of the component.<br/>"
    "Note: there are special named components "
    "with different protection types and scopes.<br/>"
    "Most prominent: \"index\" for authentication"
)


_required_passes_help = _(
    "How many protection must be passed? "
    "Set greater 0 to enable protection based access"
)

_feature_help = _(
    "Appears as featured on \"home\" page"
)


class UserComponentManager(models.Manager):
    def _update_args(self, defaults, kwargs):
        if defaults is None:
            defaults = {}
        if kwargs is None:
            kwargs = defaults
        name = kwargs.get("name", defaults.get("name", None))
        if name == "index" and force_captcha:
            defaults["required_passes"] = 2
            defaults["strength"] = 10
        elif name == "index":
            defaults["required_passes"] = 1
            defaults["strength"] = 10
        elif kwargs.get("public", defaults.get("public", False)):
            if kwargs.get(
                "required_passes",
                defaults.get("required_passes", 0)
            ) == 0:
                defaults["strength"] = 0
            else:
                defaults["strength"] = 4
        else:
            if kwargs.get(
                "required_passes",
                defaults.get("required_passes", 0)
            ) == 0:
                defaults["strength"] = 5
            else:
                defaults["strength"] = 9
        return defaults

    def create(self, **kwargs):
        ret = self.get_queryset().create(
            **self._update_args(kwargs, None)
        )
        if not ret.token:
            ret.token = create_b64_id_token(ret.id, "_")
            ret.save(update_fields=["token"])
        return ret

    def update_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().update_or_create(
            defaults=self._update_args(defaults, kwargs), **kwargs
        )
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def get_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().get_or_create(
            defaults=self._update_args(defaults, kwargs), **kwargs
        )
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def from_url_part(self, url):
        """ can be full url or token/accessmethod """
        res = static_token_matcher.match(url)
        if not res:
            raise self.model.DoesNotExist()
        res = res.groupdict()
        if res["access"] == "list":
            return self.get(
                token=res["static_token"]
            )
        else:
            return self.get(
                contents__token=res["static_token"]
            )


class UserComponent(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    # brute force protection and identifier, replaces nonce
    #  16 = usercomponent.id in hexadecimal
    #  +1 for seperator
    token = models.CharField(
        max_length=(MAX_TOKEN_B64_SIZE)+hex_size_of_bigid+2,
        db_index=True, unique=True, null=True, blank=True,
        validators=[
            validator_token
        ]
    )
    public = models.BooleanField(
        default=False,
        help_text=_(
            "Is public? Is listed and searchable?<br/>"
            "Note: This field is maybe not deactivatable "
            "because of assigned content"
        )
    )
    # special name: index:
    #    protections are used for authentication
    #    attached content is only visible for admin and user
    # db_index=True: "index" requests can speed up
    # regex disables controlcars and disable special spaces
    name = models.CharField(
        max_length=255,
        null=False,
        db_index=True,
        help_text=_name_help,
        validators=[validators.RegexValidator(r"^(\w[\w ]*\w|\w?)$")]
    )
    description = models.TextField(
        default="",
        help_text=_(
            "Description of user component."
        ), blank=True
    )
    required_passes = models.PositiveIntegerField(
        default=0,
        help_text=_required_passes_help
    )
    # cached protection strength
    strength = models.PositiveSmallIntegerField(
        default=0,
        validators=[validators.MaxValueValidator(10)],
        editable=False
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
    )
    features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="feature_for_components", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.component_feature.value
        )
    )
    default_content_features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="default_feature_for_contents", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.content_feature.value
        ),
        help_text=_(
            "Select features used by default for contents in this "
            "component"
        )
    )
    primary_anchor = models.ForeignKey(
        "spider_base.AssignedContent", related_name="primary_anchor_for",
        null=True, blank=True,
        limit_choices_to={
            "info__contains": "\x1eanchor\x1e",
        }, on_delete=models.SET_NULL,
        help_text=_(
            "Select main identifying anchor. Also used for attaching "
            "persisting tokens (elsewise they are attached to component)"
        )
    )
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    # only admin
    featured = models.BooleanField(default=False, help_text=_feature_help)

    # both should not be edited
    can_auth = models.BooleanField(default=False, editable=False)

    token_duration = models.DurationField(
        default=default_uctoken_duration,
        null=False
    )
    # only editable for admins
    deletion_requested = models.DateTimeField(
        null=True, default=None, blank=True
    )
    # fix linter warning
    objects = UserComponentManager()
    contents = None
    # should be used for retrieving active protections, related_name
    protections = None

    class Meta:
        unique_together = [("user", "name")]
        permissions = [("can_feature", "Can feature User Components")]

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<UserComponent: (%s: %s)>" % (self.username, self.__str__())

    def get_component_quota(self):
        return get_settings_func(
            "SPIDER_GET_QUOTA",
            "spkcspider.apps.spider.functions.get_quota"
        )(self.user, "usercomponents")

    def clean(self):
        _ = gettext
        self.public = (self.public and self.is_public_allowed)
        self.featured = (self.featured and self.public)
        assert(self.is_index or self.strength < 10)

        obj = self.contents.filter(
            strength__gt=self.strength
        ).order_by("strength").last()
        if obj:
            raise ValidationError(
                _(
                    'Protection strength too low, required: %(strength)s'
                ),
                code="strength",
                params={'strength': obj.strength},
            )

    def auth(self, request, ptype=ProtectionType.access_control.value,
             protection_codes=None, **kwargs):
        AssignedProtection = apps.get_model("spider_base.AssignedProtection")
        return AssignedProtection.authall(
            request, self, ptype=ptype, protection_codes=protection_codes,
            **kwargs
        )

    def get_accumulated_size(self):
        _local_size = 0
        _remote_size = 0
        for elem in self.contents.all():
            if VariantType.component_feature.value in elem.ctype.ctype:
                _remote_size += elem.get_size()
            else:
                _local_size += elem.get_size()
        return _local_size, _remote_size

    def get_absolute_url(self):
        return reverse(
            "spider_base:ucontent-list",
            kwargs={
                "token": self.token
            }
        )

    @property
    def untrusted_strength(self):
        if self.strength in {0, 4, 5, 9, 10}:
            return self.strength
        elif self.strength < 4:
            return 3
        elif self.strength < 9:
            return 8
        else:
            logger.warning("Invalid strength: %s", self.strength)

    @property
    def username(self):
        return getattr(self.user, self.user.USERNAME_FIELD)

    @property
    def user_info(self):
        return self.user.spider_info

    @property
    def index(self):
        return UserComponent.objects.get(user=self.user, name="index")

    @property
    def is_index(self):
        return (self.name == "index")

    @property
    def is_public_allowed(self):
        """ Is the public attribute allowed """
        return not self.is_index and not self.contents.filter(
            strength__gte=5
        ).exists()

    @property
    def deletion_period(self):
        return getattr(
            settings, "SPIDER_COMPONENTS_DELETION_PERIODS", {}
        ).get(self.__str__(), None)


class UserInfo(models.Model):
    """ Contains generated Informations about user """
    id = models.BigAutoField(primary_key=True, editable=False)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, editable=False,
        related_name="spider_info",
    )
    allowed_content = models.ManyToManyField(
        "spider_base.ContentVariant", related_name="+", editable=False
    )
    used_space_local = models.BigIntegerField(default=0, editable=False)
    used_space_remote = models.BigIntegerField(default=0, editable=False)

    class Meta:
        default_permissions = ()

    def calculate_allowed_content(self):
        from ..contents import installed_contents
        ContentVariant = apps.get_model("spider_base.ContentVariant")
        allowed = []
        cfilterfunc = get_settings_func(
            "SPIDER_CONTENTVARIANT_FILTER",
            "spkcspider.apps.spider.functions.allow_all_filter"
        )
        # Content types which are not "installed" should be removed/never used
        # unlisted can be removed as sideproduct if not specified with feature
        for variant in ContentVariant.objects.exclude(
            ~models.Q(ctype__contains=VariantType.content_feature.value) |
            ~models.Q(ctype__contains=VariantType.component_feature.value),
            ctype__contains=VariantType.unlisted.value
        ).filter(
            code__in=installed_contents
        ):
            if cfilterfunc(self.user, variant):
                allowed.append(variant)
        # save not required, m2m field
        self.allowed_content.set(allowed)

    def calculate_used_space(self):
        from . import AssignedContent
        self.used_space_local = 0
        self.used_space_remote = 0
        for c in AssignedContent.objects.filter(
            usercomponent__user=self.user
        ):
            if VariantType.component_feature.value in c.ctype.ctype:
                self.used_space_remote += c.get_size()
            else:
                self.used_space_local += c.get_size()

    def update_with_quota(self, size_diff, quota_type):
        fname = "used_space_{}".format(quota_type)
        qval = getattr(self, fname)
        quota = get_settings_func(
            "SPIDER_GET_QUOTA",
            "spkcspider.apps.spider.functions.get_quota"
        )(self.user, quota_type)
        # if over quota: reducing size is always good and should never fail
        if quota and size_diff > 0 and qval + size_diff > quota:
            raise ValidationError(
                _("Exceeds quota by %(diff)s Bytes"),
                code='quota_exceeded',
                params={'diff': size_diff},
            )
        setattr(
            self, fname, models.F(fname)+size_diff
        )



# Generated by Django 2.2 on 2019-04-07 14:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('spider_base', '0013_auto_20190407_1313'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='assignedcontent',
            unique_together={('content_type', 'object_id')},
        ),
        migrations.AddConstraint(
            model_name='assignedcontent',
            constraint=models.UniqueConstraint(fields=('usercomponent', 'info'), name='unique_info'),
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-30 16:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0006_verifysourceobject_update_secret'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='verifysourceobject',
            name='update_secret',
        ),
        migrations.AddField(
            model_name='verifysourceobject',
            name='token',
            field=models.CharField(blank=True, max_length=126, null=True, unique=True),
        ),
    ]



__all__ = [
    "LicenseChooserWidget"
]


import json

# from django.forms import widgets
from django.conf import settings
# from django.utils.translation import gettext_lazy as _

from spkcspider.apps.spider.widgets import OpenChoiceWidget


_extra = '' if settings.DEBUG else '.min'


class LicenseChooserWidget(OpenChoiceWidget):
    licenses = None

    def __init__(self, licenses, allow_multiple_selected=False, **kwargs):
        self.licenses = licenses
        super().__init__(
            allow_multiple_selected=allow_multiple_selected, **kwargs
        )

    def build_attrs(self, base_attrs, extra_attrs=None):
        """ add license_urls to attrs."""
        ret = super().build_attrs(base_attrs, extra_attrs)
        d = dict(map(lambda x: (x[0], x[1]["url"]), self.licenses.items()))
        ret["license_urls"] = json.dumps(d)
        # ret.setdefault("style", "color:black;")
        return ret



# Generated by Django 2.1.5 on 2019-01-20 12:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_webcfg', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='webconfig',
            name='creation_url',
        ),
        migrations.RemoveField(
            model_name='webconfig',
            name='token',
        ),
    ]



__all__ = [
    "clean_graph", "get_hashob", "validate_request_default",
    "verify_tag_default", "domain_auth"
]

import logging
from urllib.parse import parse_qs, urlencode

from django.conf import settings
from django.urls import reverse
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.test import Client
from django.http.request import validate_host

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from rdflib import XSD, URIRef, Literal

import requests

from spkcspider.apps.spider.constants import spkcgraph, host_tld_matcher


def validate_request_default(request, form):
    if not form.is_valid():
        return False

    return True


def verify_tag_default(tag):
    tag.save()


def domain_auth(source, hostpart):
    GET = parse_qs(source.get_params)
    if "token" not in GET:
        GET["token"] = "prefer"
    GET["intention"] = "domain"
    GET["referrer"] = "{}{}".format(
        hostpart,
        reverse("spider_verifier:create")
    )
    source.initialize_token()
    source.save(update_fields=["token"])
    GET["payload"] = urlencode(
        {
            "url": source.url,
            "update_secret": source.token
        }
    )

    url = "{}?{}".format(
        source.url, urlencode(GET)
    )
    ret = True
    params, can_inline = get_requests_params(url)

    if can_inline:
        response = Client().get(
            url, follow=True, secure=True, Connection="close"
        )
        if response.status_code >= 400:
            ret = False
    else:
        try:
            with requests.get(
                url, headers={"Connection": "close"},
                **params
            ) as resp:
                resp.close()
                resp.raise_for_status()
        except Exception:
            if settings.DEBUG:
                logging.exception("domain_auth failed")
            ret = False
    # other or own update_secret was successful or url has problems
    source.refresh_from_db()
    return ret


def clean_graph(mtype, graph, start, source, hostpart):
    if not mtype:
        return None
    elif "UserComponent" in mtype:
        return "list"
    elif "SpiderTag" in mtype:
        if (
            URIRef(start),
            spkcgraph["features"],
            Literal("verify", datatype=XSD.string)
        ) in graph:
            return "layout_cb"
        ret = "layout"
        if domain_auth(source, hostpart):
            graph.set(
                (
                    start,
                    spkcgraph["action:view"],
                    Literal(source.get_url(), datatype=XSD.anyURI)
                )
            )
            ret = "layout_cb"
        return ret
    else:
        return "content"


def get_hashob():
    return hashes.Hash(
        getattr(
           settings, "VERIFICATION_HASH_ALGORITHM",
           settings.SPIDER_HASH_ALGORITHM
        ),
        backend=default_backend()
    )


def get_requests_params(url):
    _url = host_tld_matcher.match(url)
    if not _url:
        raise ValidationError(
            _("Invalid URL: \"%(url)s\""),
            code="invalid_url",
            params={"url": url}
        )
    _url = _url.groupdict()
    mapper = getattr(
        settings, "VERIFIER_REQUEST_KWARGS_MAP",
        settings.SPIDER_REQUEST_KWARGS_MAP
    )
    return (
        mapper.get(
            _url["host"],
            mapper.get(
                _url["tld"],  # maybe None but then fall to retrieval 3
                mapper[b"default"]
            )
        ),
        not getattr(settings, "SPIDER_DISABLE_FAKE_CLIENT", None) and
        validate_host(_url["host"], settings.ALLOWED_HOSTS)
    )



__all__ = [
    "LinkForm", "TravelProtectionForm", "TravelProtectionManagementForm"
]

from datetime import timedelta
from base64 import b64encode

from django import forms
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.contrib.auth import authenticate

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

from ..constants import (
    dangerous_login_choices, travel_scrypt_params, TravelLoginType,
    loggedin_active_tprotections, VariantType
)
from ..models import LinkContent, TravelProtection, AssignedContent
from ..fields import MultipleOpenChoiceField, ContentMultipleChoiceField
from ..helpers import get_settings_func
from ..widgets import (
    OpenChoiceWidget, RomeDatetimePickerWidget, Select2Widget
)


_extra = '' if settings.DEBUG else '.min'


_time_help_text = _(
    "Time in \"{}\" timezone"
)


_login_protection = _(
    "Hide: Hide protected contents and components<br/>"
    "Hide if triggered: Hide protected contents and components if triggered<br/>"  # noqa: E501
    "Disable: Disable Login (not available on Self-Protection (useless))<br/>"
    "Disable if triggered: Hide protected contents and components and disable login if triggered<br/>"  # noqa: E501
    "Wipe: Wipe protected content on login (maybe not available)<br/>"
    "Wipe User: destroy user on login (maybe not available)<br/><br/>"
    "Note: login protections work only on login. Except hide and disable protections, which works also for logged in users<br/>"  # noqa: E501
    "Note: the hide and disable protections disable also admin access when active (elsewise easy circumventable)"  # noqa: E501
)


class LinkForm(forms.ModelForm):

    class Meta:
        model = LinkContent
        fields = ['content', 'push']

    def __init__(self, uc, request, **kwargs):
        super().__init__(**kwargs)
        # if self.instance.associated:
        #     if "\x1eanchor\x1e" in self.instance.associated:
        #         self.fields["content"].disabled = True
        q = self.fields["content"].queryset
        travel = TravelProtection.objects.get_active_for_request(request)
        travel = travel.filter(
            login_protection__in=loggedin_active_tprotections
        )
        self.fields["content"].queryset = q.filter(
            strength__lte=uc.strength
        ).exclude(usercomponent__travel_protected__in=travel)
        # component auth should limit links to visible content
        # read access outside from component elsewise possible
        if request.user != uc.user and not request.is_staff:
            q = self.fields["content"].queryset
            self.fields["content"].queryset = q.filter(
                models.Q(usercomponent=uc) |
                models.Q(referenced_by__usercomponent=uc)
            )


class TravelProtectionManagementForm(forms.ModelForm):
    class Meta:
        model = TravelProtection
        fields = []

    def clean(self):
        super().clean()
        pwset = set(self.instance.associated.getlist("pwhash", 20))
        self.instance._encoded_form_info = "".join(
            map(lambda x: "pwhash={}\x1e".format(x), pwset)
        )
        if self.instance.associated.getflag("anonymous_deactivation"):
            self.instance._encoded_form_info = \
                "{}anonymous_deactivation\x1e".format(
                    self.instance._encoded_form_info
                )
        if self.instance.associated.getflag("anonymous_trigger"):
            self.instance._encoded_form_info = \
                "{}anonymous_trigger\x1e".format(
                    self.instance._encoded_form_info
                )

        return self.cleaned_data


class TravelProtectionForm(forms.ModelForm):
    uc = None
    # self_protection = forms.ChoiceField(
    #    label=_("Self-protection"), help_text=_(_self_protection),
    #     initial="None", choices=PROTECTION_CHOICES
    # )
    start = forms.DateTimeField(
        required=True, widget=RomeDatetimePickerWidget(),
        help_text=_time_help_text
    )
    stop = forms.DateTimeField(
        required=False, widget=RomeDatetimePickerWidget(),
        help_text=_time_help_text
    )
    trigger_pws = MultipleOpenChoiceField(
        label=_("Trigger Passwords"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    approved = forms.BooleanField(disabled=True)
    overwrite_pws = forms.BooleanField(
        required=False, initial=False,
        help_text=_(
            "If set, the set of passwords is overwritten or "
            "removed (if no passwords are set)"
        )
    )
    anonymous_deactivation = forms.BooleanField(
        required=False, initial=False,
        help_text=_(
            "Can deactivate protection without login"
        )
    )
    protect_contents = ContentMultipleChoiceField(
        queryset=AssignedContent.objects.all(), required=False,
        to_field_name="id",
        widget=Select2Widget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    master_pw = forms.CharField(
        label=_("Master Login Password"),
        widget=forms.PasswordInput(render_value=True), required=True
    )

    class Meta:
        model = TravelProtection
        fields = [
            "active", "start", "stop", "login_protection",
            "protect_components", "protect_contents", "trigger_pws"
        ]
        widgets = {
            "protect_components": Select2Widget(
                 allow_multiple_selected=True,
                 attrs={
                     "style": "min-width: 250px; width:100%"
                 }
            )
        }
        help_texts = {
            "login_protection": _login_protection,
            "master_pw": _(
                "Enter Password used for the user account"
            )
        }

    class Media:
        js = []

    @classmethod
    def _filter_selfprotection(cls, x):
        if x[0] == TravelLoginType.disable.value:
            return False
        return True

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        if self.instance.associated.ctype.name == "TravelProtection":
            self.fields["start"].help_text = str(
                self.fields["start"].help_text
            ).format(timezone.get_current_timezone_name())
            self.fields["stop"].help_text = str(
                self.fields["stop"].help_text
            ).format(timezone.get_current_timezone_name())
            if not getattr(self.instance, "id"):
                now = timezone.now()
                self.initial["start"] = now + timedelta(hours=3)
                self.initial["stop"] = now + timedelta(days=7)
        else:
            self.fields["login_protection"].choices = \
                filter(
                    self._filter_selfprotection,
                    self.fields["login_protection"].choices
                )
            del self.fields["start"]
            del self.fields["stop"]
        if not getattr(self.instance, "id") and not self.data:
            self.initial["active"] = True

        travel = TravelProtection.objects.get_active_for_request(
            request
        ).filter(login_protection__in=loggedin_active_tprotections)
        selfid = getattr(self.instance, "id", -1)
        if self.initial["login_protection"] not in dangerous_login_choices:
            del self.fields["approved"]
        if self.initial["login_protection"] == TravelLoginType.disable.value:
            self.initial["login_protection"] = \
                TravelLoginType.trigger_disable.value

        q_component = models.Q(
            user=request.user
        ) & (
            ~models.Q(travel_protected__in=travel) |
            models.Q(travel_protected__id=selfid)
        )
        self.fields["protect_components"].queryset = \
            self.fields["protect_components"].queryset.filter(
                q_component
            ).distinct().order_by(
                "name"
            )

        q_content = models.Q(usercomponent__user=request.user) & (
            ~(
                models.Q(usercomponent__travel_protected__in=travel) |
                models.Q(travel_protected__in=travel) |
                # don't allow detectable contents
                models.Q(info__contains="\x1eanchor\x1e") |
                models.Q(info__contains="\x1eprimary\x1e") |
                # contents also appearing as features are easily detectable
                models.Q(
                    ctype__ctype__contains=VariantType.feature_connect.value
                )
            ) |
            (
                models.Q(usercomponent__travel_protected__id=selfid) |
                models.Q(travel_protected__id=selfid)
            )
        )
        if getattr(self.instance, "id", None):
            # exlude own content
            q_content &= ~models.Q(pk=self.instance.associated.pk)
            self.initial["anonymous_deactivation"] = \
                self.instance.associated.getflag("anonymous_deactivation")
            if self.instance.associated.getlist("pwhash", 1):
                self.fields["trigger_pws"].help_text = _(
                    "<span style='color:red'>Trigger passwords already set."
                    "</span>"
                )
            else:
                self.fields["trigger_pws"].help_text = _(
                    "No trigger passwords set."
                )
                del self.fields["overwrite_pws"]
        if (
            not self.instance.associated.getlist("pwhash", 1) and
            self.instance.associated.ctype.name == "SelfProtection"
        ):
            self.fields["trigger_pws"].required = True
        # use q for filtering (including own)
        self.fields["protect_contents"].queryset = \
            self.fields["protect_contents"].queryset.filter(
                q_content
            ).distinct().order_by(
                "name"
            )

    def clean(self):
        super().clean()
        self.initial["master_pw"] = self.cleaned_data.pop("master_pw", None)
        if not authenticate(
            self.request,
            username=self.instance.associated.usercomponent.username,
            password=self.initial.get("master_pw", None),
            nospider=True
        ):
            self.add_error(
                "master_pw",
                forms.ValidationError(
                    _("Invalid Password"),
                    code="invalid_password"
                )
            )
        if "trigger_pws" in self.cleaned_data:
            pwset = set(map(
                lambda x: b64encode(Scrypt(
                    salt=settings.SECRET_KEY.encode("utf-8"),
                    backend=default_backend(),
                    **travel_scrypt_params
                ).derive(x[:128].encode("utf-8"))).decode("ascii"),
                self.cleaned_data["trigger_pws"]
            ))
            if (
                self.instance.associated.ctype.name == "SelfProtection" and
                self.cleaned_data.get("overwrite_pws") and
                len(pwset) == 0
            ):
                self.add_error("trigger_pws", forms.ValidationError(
                    _(
                        "Clearing trigger passwords not allowed "
                        "(for Self-Protection)"
                    )
                ))
            if not self.cleaned_data.get("overwrite_pws") and len(pwset) < 20:
                pwset.update(self.instance.associated.getlist(
                    "pwhash", 20-len(pwset)
                ))
        else:
            pwset = set(self.instance.associated.getlist("pwhash", 20))
        self.instance._encoded_form_info = "".join(
            map(lambda x: "pwhash={}\x1e".format(x), pwset)
        )
        if self.cleaned_data.get("start") and self.cleaned_data.get("stop"):
            if self.cleaned_data["stop"] < self.cleaned_data["start"]:
                self.add_error("start", forms.ValidationError(
                    _("Stop < Start")
                ))
        if self.cleaned_data.get("anonymous_deactivation"):
            self.instance._encoded_form_info = \
                "{}anonymous_deactivation\x1e".format(
                    self.instance._encoded_form_info
                )
        if self.cleaned_data.get("anonymous_trigger"):
            self.instance._encoded_form_info = \
                "{}anonymous_trigger\x1e".format(
                    self.instance._encoded_form_info
                )
        if self.cleaned_data["login_protection"] in dangerous_login_choices:
            self.cleaned_data["approved"] = get_settings_func(
                "SPIDER_DANGEROUS_APPROVE",
                "spkcspider.apps.spider.functions.approve_dangerous"
            )(self)
            if self.cleaned_data["approved"] is False:
                self.add_error("login_protection", forms.ValidationError(
                    _("This login protection is not allowed")
                ))
        else:
            self.cleaned_data["approved"] = True
        try:
            self.cleaned_data["protect_contents"] = \
                self.cleaned_data["protect_contents"].exclude(
                    usercomponent__in=self.cleaned_data["protect_components"]
                )
        except KeyError:
            pass
        return self.cleaned_data

    def is_valid(self):
        isvalid = super().is_valid()
        if not getattr(self, "cleaned_data", None):
            return False
        return isvalid

    def _save_m2m(self):
        super()._save_m2m()
        self.instance.protect_contents.add(self.instance.associated)

    def save(self, commit=True):
        self.instance.approved = bool(self.cleaned_data["approved"])
        return super().save(commit)



# Generated by Django 2.1.7 on 2019-02-21 22:53

from django.db import migrations, models
import django.db.models.deletion
import spkcspider.apps.spider_keys.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0003_auto_20190130_1137'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='publickey',
            name='use_for_encryption',
        ),
        migrations.AlterField(
            model_name='anchorkey',
            name='key',
            field=models.OneToOneField(help_text='"Public Key"-Content for signing identifier. It is recommended to use different keys for signing and encryption.', on_delete=django.db.models.deletion.CASCADE, related_name='anchorkey', to='spider_keys.PublicKey'),
        ),
        migrations.AlterField(
            model_name='publickey',
            name='key',
            field=models.TextField(help_text='It is recommended to use different keysfor signing and encryption', validators=[spkcspider.apps.spider_keys.models.valid_pkey_properties]),
        ),
    ]



__all__ = ["BaseInfoModel", "ReferrerObject", "info_and", "info_or"]

import re

from django.conf import settings
from django.db import models
from django.core.exceptions import ValidationError
from django.utils.translation import gettext
from django.utils.functional import cached_property

from ..helpers import extract_host

_info_replacer_templ = '\x1e{}.*\x1e'


def info_and(*args, **kwargs):
    q = models.Q()
    for query in args:
        if isinstance(query, models.Q):
            q &= query
        else:
            q &= models.Q(
                info__contains="\x1e{}\x1e".format(query)
            )
    for name, query in kwargs.items():
        if query is None:
            q &= models.Q(
                info__contains="\x1e{}=".format(name)
            )
        else:
            q &= models.Q(
                info__contains="\x1e{}={}\x1e".format(name, query)
            )
    return q


def info_or(*args, **kwargs):
    q = models.Q()
    for query in args:
        if isinstance(query, models.Q):
            q |= query
        else:
            q |= models.Q(
                info__contains="\x1e{}\x1e".format(query)
            )
    for name, query in kwargs.items():
        if query is None:
            q |= models.Q(
                info__contains="\x1e{}=".format(name)
            )
        else:
            q |= models.Q(
                info__contains="\x1e{}={}\x1e".format(name, query)
            )
    return q


def info_field_validator(value):
    _ = gettext
    if value[-1] != "\x1e":
        raise ValidationError(
            _('%(value)s ends not with "\\x1e"'),
            code="syntax",
            params={'value': value},
        )
    if value[0] != "\x1e":
        raise ValidationError(
            _('%(value)s starts not with "\\x1e"'),
            code="syntax",
            params={'value': value},
        )
    # check elements
    for elem in value.split("\x1e"):
        if elem == "":
            continue
        f = elem.find("=")
        # no flag => allow multiple instances
        if f != -1:
            continue
        counts = 0
        counts += value.count("\x1e%s\x1e" % elem)
        # check: is flag used as key in key, value storage
        counts += value.count("\x1e%s=" % elem)
        assert counts > 0, value
        if counts > 1:
            raise ValidationError(
                _('flag not unique: %(element)s in %(value)s'),
                params={'element': elem, 'value': value},
            )


class ReferrerObject(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    url = models.URLField(
        max_length=(
            600 if (
                settings.DATABASES["default"]["ENGINE"] !=
                "django.db.backends.mysql"
            ) else 255
        ),
        db_index=True, unique=True, editable=False
    )

    @cached_property
    def host(self):
        return extract_host(self.url)


class BaseInfoModel(models.Model):

    class Meta:
        abstract = True

    # for extra information over content, admin only editing
    # format: \nflag1\nflag2\nfoo=true\nfoo2=xd\n...\nendfoo=xy\n
    # every section must start and end with \n every keyword must be unique and
    # in this format: keyword=
    # no unneccessary spaces!
    info = models.TextField(
        null=False, editable=False,
        validators=[info_field_validator]
    )

    def getflag(self, flag):
        if "\x1e%s\x1e" % flag in self.info:
            return True
        return False

    def getlist(self, key, amount=None):
        info = self.info
        ret = []
        pstart = info.find("\x1e%s=" % key)
        while pstart != -1:
            tmpstart = pstart+len(key)+2
            pend = info.find("\x1e", tmpstart)
            if pend == -1:
                raise Exception(
                    "Info field error: doesn't end with \"\\x1e\": \"%s\"" %
                    info
                )
            ret.append(info[tmpstart:pend])
            pstart = info.find("\x1e%s=" % key, pend)
            # if amount=0 => bool(amount) == false
            if amount and amount <= len(ret):
                break
        return ret

    def replace_info(self, **kwargs):
        """
            Warning! Order dependend (especially for unique tests)
        """
        rep_replace = []
        rep_missing = []
        self.info = self.info.replace("{}", "{{}}")
        for name, val in kwargs.items():
            pattern = re.compile(
                _info_replacer_templ.format(re.escape(name)), re.M
            )
            if not val:
                # remove name
                self.info = pattern.sub(
                    "\x1e", self.info, 0
                )
                continue
            # count replacements
            self.info, count1 = pattern.subn(
                "\x1e{}\x1e", self.info, 1
            )
            if count1 == 0:
                rep = rep_missing
            else:
                rep = rep_replace
                self.info = pattern.sub(
                    "\x1e", self.info, 0
                )
            if val is True:
                rep.append(name)
            elif isinstance(val, (tuple, list)):
                rep.append("\x1e".join(
                    map(
                        lambda x: "{}={}".format(name, x),
                        val
                    )
                ))
            else:
                rep.append("{}={}".format(name, val))
        self.info = self.info.format(*rep_replace)
        if rep_missing:
            self.info = "{}\x1e{}\x1e".format(
                self.info, "\x1e".join(rep_missing)
            )
        return rep_missing

    def info_and(self, *args, **kwargs):
        return info_and(*args, **kwargs)

    def info_or(self, *args, **kwargs):
        return info_or(*args, **kwargs)



# Generated by Django 2.1.5 on 2019-01-30 11:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0002_anchorkey_anchorserver'),
    ]

    operations = [
        migrations.AddField(
            model_name='publickey',
            name='use_for_encryption',
            field=models.BooleanField(blank=True, default=True, help_text='Use for encryption not for signing.'),
        ),
        migrations.AlterField(
            model_name='anchorkey',
            name='key',
            field=models.OneToOneField(help_text='"Public Key"-Content', limit_choices_to={'use_for_encryption': False}, on_delete=django.db.models.deletion.CASCADE, related_name='anchorkey', to='spider_keys.PublicKey'),
        ),
    ]



# Generated by Django 2.2 on 2019-04-05 00:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0010_auto_20190403_0417'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='filefilet',
            name='name',
        ),
        migrations.RemoveField(
            model_name='textfilet',
            name='name',
        ),
        migrations.RemoveField(
            model_name='textfilet',
            name='preview_words',
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-04 16:40

from django.db import migrations


def move_tokens_forward(apps, schema_editor):
    AssignedContent = apps.get_model('spider_base', 'AssignedContent')
    UserComponent = apps.get_model('spider_base', 'UserComponent')
    for row in AssignedContent.objects.all():
        if row.token:
            row.token = '_'.join(row.token.split("/"))
        row.save()
    for row in UserComponent.objects.all():
        if row.token:
            row.token = '_'.join(row.token.split("/"))
        row.save()


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0016_auto_20190414_1743'),
    ]

    operations = [
        migrations.RunPython(move_tokens_forward),
    ]
